
/** @file fourvector.h
 *
 * A fourvector class
 *
 */

/*
 *  Copyright (C) 2011 Stefan Weinzierl
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __SISCONE_PARTON_FOURVECTOR_H__
#define __SISCONE_PARTON_FOURVECTOR_H__

#include <cmath>

/**
 *
 * The class fourvector
 *
 */
class fourvector {

  // ctors
 public :  
  fourvector(void);
  fourvector(double t, double x, double y, double z);

  // functions
 public :
  fourvector & sum_up(const fourvector & v);
  fourvector & rescale(double c);

  double transverse_momentum(void) const;
  double rapidity(void) const;
  double azimuthal_angle(void) const;

  double spatial_norm(void) const;
  double spatial_scalar_product(const fourvector & q) const;
  double energy_component(void) const;

  // member variables :
 protected : 
  double t;
  double x;
  double y;
  double z;
};

// ----------------------------------------------------------------------------

/**
 *
 * Default constructor
 *
 */
inline fourvector::fourvector(void) : t(0.0), x(0.0), y(0.0), z(0.0)
{}

/**
 *
 * Standard constructor
 *
 */
inline fourvector::fourvector(double tt, double xx, double yy, double zz) :
		  t(tt), x(xx), y(yy), z(zz)
{}

/**
 *
 * Adds \f$v\f$ to the fourvector.
 *
 */
inline fourvector & fourvector::sum_up(const fourvector & v)
{
  t += v.t;
  x += v.x;
  y += v.y;
  z += v.z;

  return *this;
}

/**
 *
 * Scales all components by \f$c\f$.
 *
 */
inline fourvector & fourvector::rescale(double c)
{
  t *= c;
  x *= c;
  y *= c;
  z *= c;

  return *this;
}

/**
 *
 * Returns the transverse momentum with respect to the
 * \f$z\f$-axis.
 *
 */
inline double fourvector::transverse_momentum(void) const
{
  // beam along z-axis
  double temp = x*x + y*y;
  if (temp>0) return sqrt(temp);
  return(0.0);
}

/**
 *
 * Returns the rapidity with respect to the
 * \f$z\f$-axis.
 *
 */
inline double fourvector::rapidity(void) const
{
  // beam along z-axis
  return 0.5*log( (t+z)/(t-z) );
}

/**
 *
 * Returns the azimuthal angle \f$\phi\f$.
 *
 */
inline double fourvector::azimuthal_angle(void) const
{
  // beam along z-axis
  return atan2(y,x);
}

/**
 *
 * The norm of the spatial components of the fourvector:
 * \f[
 *     v_x^2 + v_y^2 + v_z^2
 * \f]
 *
 */
inline double fourvector::spatial_norm(void) const
{
  return x*x + y*y + z*z;
}

/**
 *
 * Returns
 * \f[
 *    \vec{p} \cdot \vec{q}
 * \f]
 *
 */
inline double fourvector::spatial_scalar_product(const fourvector & q) const
{
  return x*q.x + y*q.y + z*q.z;
}

/**
 *
 * Returns the energy component.
 *
 */
inline double fourvector::energy_component(void) const
{
  return t;
}

#endif // ndef __SISCONE_PARTON_FOURVECTOR_H__

